@@ -5,7 +5,7 @@
 
    Distributed under the GPL v2
    see the file COPYING for details
-   or visit http://www.gnu.org/copyleft/gpl.html
+   or visit https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 
    \file
 
@@ -15,11 +15,16 @@
 
 #include "common/common_pch.h"
 
-#include <boost/date_time/posix_time/posix_time.hpp>
 #include <cmath>
 #include <iostream>
+#if defined(SYS_UNIX) || defined(SYS_APPLE)
+# include <signal.h>
+#endif
 #include <typeinfo>
 
+#include <QDateTime>
+#include <QRegularExpression>
+
 #include <ebml/EbmlHead.h>
 #include <ebml/EbmlSubHead.h>
 #include <ebml/EbmlVersion.h>
@@ -61,6 +66,8 @@
 #include "common/mm_null_io.h"
 #include "common/mm_proxy_io.h"
 #include "common/mm_write_buffer_io.h"
+#include "common/path.h"
+#include "common/qt.h"
 #include "common/strings/formatting.h"
 #include "common/tags/tags.h"
 #include "common/translation.h"
@@ -114,13 +121,12 @@ bool g_no_linking                                             = true;
 bool g_use_durations                                          = false;
 bool g_no_track_statistics_tags                               = false;
 bool g_write_date                                             = true;
+bool g_stop_after_video_ends                                  = false;
 
 double g_timestamp_scale                                      = TIMESTAMP_SCALE;
 timestamp_scale_mode_e g_timestamp_scale_mode                 = timestamp_scale_mode_e{TIMESTAMP_SCALE_MODE_NORMAL};
 
 double g_video_fps                                            = -1.0;
-int g_default_tracks[3]                                       = { 0, 0, 0, };
-int g_default_tracks_priority[3]                              = { 0, 0, 0, };
 
 bool g_identifying                                            = false;
 identification_output_format_e g_identification_output_format = identification_output_format_e::text;
@@ -135,7 +141,7 @@ mtx::chapters::kax_cptr g_kax_chapters;
 std::unique_ptr<KaxTags> g_tags_from_cue_chapters;
 
 std::string g_chapter_file_name;
-std::string g_chapter_language;
+mtx::bcp47::language_c g_chapter_language;
 std::string g_chapter_charset;
 
 std::string g_segmentinfo_file_name;
@@ -150,14 +156,16 @@ int64_t g_tags_size                         = 0;
 int g_file_num = 1;
 
 int g_split_max_num_files                   = 65535;
-std::string g_splitting_by_chapters_arg;
+bool g_splitting_by_all_chapters            = false;
+std::unordered_map<unsigned int, int> g_splitting_by_chapter_numbers;
 
 append_mode_e g_append_mode                 = APPEND_MODE_FILE_BASED;
 bool s_appending_files                      = false;
 auto s_debug_appending                      = debugging_option_c{"append|appending"};
 auto s_debug_rerender_track_headers         = debugging_option_c{"rerender|rerender_track_headers"};
+auto s_debug_splitting_chapters             = debugging_option_c{"splitting_chapters"};
 
-std::string g_default_language              = "und";
+mtx::bcp47::language_c g_default_language;
 
 mtx::bits::value_cptr g_seguid_link_previous;
 mtx::bits::value_cptr g_seguid_link_next;
@@ -176,7 +184,7 @@ static std::unique_ptr<EbmlVoid> s_kax_chapters_void;
 static int64_t s_max_chapter_size           = 0;
 static std::unique_ptr<EbmlVoid> s_void_after_track_headers;
 
-static std::vector<std::tuple<timestamp_c, std::string, std::string>> s_additional_chapter_atoms;
+static std::vector<std::tuple<timestamp_c, std::string, mtx::bcp47::language_c>> s_additional_chapter_atoms;
 
 static mm_io_cptr s_out;
 
@@ -185,13 +193,15 @@ static mtx::bits::value_c s_seguid_prev(128), s_seguid_current(128), s_seguid_ne
 static std::unique_ptr<EbmlHead> s_head;
 
 static std::string s_muxing_app, s_writing_app;
-static boost::posix_time::ptime s_writing_date;
+static QDateTime s_writing_date;
 
-static boost::optional<int64_t> s_maximum_progress;
+static std::optional<int64_t> s_maximum_progress;
 int64_t s_current_progress{};
 
 std::unique_ptr<mtx::doc_type_version_handler_c> g_doc_type_version_handler;
 
+bool g_deterministic{}, g_use_legacy_font_mime_types{};
+
 /** \brief Add a segment family UID to the list if it doesn't exist already.
 
   \param family This segment family element is converted to a 128 bit
@@ -244,10 +254,17 @@ update_ebml_head() {
    segment duration and the segment length.
 */
 #if defined(SYS_UNIX) || defined(SYS_APPLE)
+void
+reraise_sigint() {
+  mxmsg(MXMSG_ERROR, Y("mkvmerge was interrupted by a SIGINT (Ctrl+C?)\n"));
+  signal(SIGINT, SIG_DFL);
+  kill(getpid(), SIGINT);
+}
+
 void
 sighandler(int /* signum */) {
   if (!s_out)
-    mxerror(Y("mkvmerge was interrupted by a SIGINT (Ctrl+C?)\n"));
+    reraise_sigint();
 
   mxwarn(Y("\nmkvmerge received a SIGINT (probably because the user pressed "
            "Ctrl+C). Trying to sanitize the file. If mkvmerge hangs during "
@@ -290,7 +307,7 @@ sighandler(int /* signum */) {
 
   cleanup();
 
-  mxerror(Y("mkvmerge was interrupted by a SIGINT (Ctrl+C?)\n"));
+  reraise_sigint();
 }
 #endif
 
@@ -302,7 +319,7 @@ add_to_progress(int64_t num_bytes_processed) {
 static int64_t
 get_maximum_progress() {
   if (!s_maximum_progress)
-    s_maximum_progress = boost::accumulate(g_files, 0ull, [](int64_t num, auto const &file) { return num + file->reader->get_maximum_progress(); });
+    s_maximum_progress = std::accumulate(g_files.begin(), g_files.end(), 0ull, [](int64_t num, auto const &file) { return num + file->reader->get_maximum_progress(); });
 
   return *s_maximum_progress;
 }
@@ -425,6 +442,15 @@ add_packetizer_globally(generic_packetizer_c *packetizer) {
   g_packetizers.push_back(pack);
 }
 
+void
+maybe_set_segment_title(std::string const &title) {
+  if (g_segment_title_set || title.empty())
+    return;
+
+  g_segment_title     = title;
+  g_segment_title_set = true;
+}
+
 static void
 set_timestamp_scale() {
   bool video_present          = false;
@@ -453,8 +479,11 @@ set_timestamp_scale() {
       && audio_present
       && (0 < highest_sample_rate)
       && (   !video_present
-          || (TIMESTAMP_SCALE_MODE_AUTO == g_timestamp_scale_mode)))
-    g_timestamp_scale = static_cast<int64_t>(1000000000.0 / highest_sample_rate - 1.0);
+          || (TIMESTAMP_SCALE_MODE_AUTO == g_timestamp_scale_mode))) {
+    auto new_timestamp_scale = static_cast<int64_t>(1'000'000'000.0 / highest_sample_rate - 1.0);
+    if (new_timestamp_scale >= 1)
+      g_timestamp_scale = new_timestamp_scale;
+  }
 
   g_max_ns_per_cluster = std::min<int64_t>(32700 * g_timestamp_scale, g_max_ns_per_cluster);
   GetChild<KaxTimecodeScale>(*s_kax_infos).SetValue(g_timestamp_scale);
@@ -510,6 +539,39 @@ generate_segment_uids() {
   s_seguid_next.generate_random();
 }
 
+static int
+sortable_track_type_for_file_and_track_id(track_order_t const &track_order) {
+  auto const type = g_files[track_order.file_id]->reader->m_reader_packetizers[track_order.track_id]->get_track_type();
+
+  return type == track_video    ?  1
+       : type == track_audio    ?  2
+       : type == track_subtitle ?  3
+       :                          99;
+}
+
+static bool
+track_order_less_than_for_type(track_order_t const &a,
+                               track_order_t const &b) {
+  return sortable_track_type_for_file_and_track_id(a) < sortable_track_type_for_file_and_track_id(b);
+}
+
+static void
+create_type_based_track_order() {
+  auto file_id = -1;
+
+  for (auto const &file : g_files) {
+    ++file_id;
+
+    if (file->appending)
+      continue;
+
+    for (int track_id = 0, end = file.get()->reader->m_reader_packetizers.size(); track_id < end; ++track_id)
+      g_track_order.push_back(track_order_t{ file_id, track_id });
+  }
+
+  std::stable_sort(g_track_order.begin(), g_track_order.end(), track_order_less_than_for_type);
+}
+
 /** \brief Render the basic EBML and Matroska headers
 
    Renders the segment information and track headers. Also reserves
@@ -521,6 +583,12 @@ render_headers(mm_io_c *out) {
   try {
     render_ebml_head(out);
 
+    if (debugging_c::requested("void_before_segment")) {
+      EbmlVoid v;
+      v.SetSize(4);
+      v.Render(*out);
+    }
+
     s_kax_infos = std::make_unique<KaxInfo>();
 
     s_kax_duration = new KaxMyDuration{ !g_video_packetizer || (TIMESTAMP_SCALE_MODE_AUTO == g_timestamp_scale_mode) ? EbmlFloat::FLOAT_64 : EbmlFloat::FLOAT_32};
@@ -539,7 +607,7 @@ render_headers(mm_io_c *out) {
     GetChild<KaxWritingApp>(*s_kax_infos).SetValueUTF8(s_writing_app);
 
     if (g_write_date)
-      GetChild<KaxDateUTC>(*s_kax_infos).SetEpochDate(s_writing_date.is_not_a_date_time() ? 0 : mtx::date_time::to_time_t(s_writing_date));
+      GetChild<KaxDateUTC>(*s_kax_infos).SetEpochDate(s_writing_date.toSecsSinceEpoch());
     else
       DeleteChildren<KaxDateUTC>(*s_kax_infos);
 
@@ -616,20 +684,22 @@ render_headers(mm_io_c *out) {
     if (first_file) {
       g_kax_last_entry = nullptr;
 
-      size_t i;
-      for (i = 0; i < g_track_order.size(); i++)
-        if ((g_track_order[i].file_id >= 0) && (g_track_order[i].file_id < static_cast<int>(g_files.size())) && !g_files[g_track_order[i].file_id]->appending)
-          g_files[g_track_order[i].file_id]->reader->set_headers_for_track(g_track_order[i].track_id);
+      if (g_track_order.empty())
+        create_type_based_track_order();
 
-      for (i = 0; i < g_files.size(); i++)
-        if (!g_files[i]->appending)
-          g_files[i]->reader->set_headers();
+      for (auto const &order : g_track_order)
+        if ((order.file_id >= 0) && (order.file_id < static_cast<int>(g_files.size())) && !g_files[order.file_id]->appending)
+          g_files[order.file_id]->reader->set_headers_for_track(order.track_id);
+
+      for (auto const &file : g_files)
+        if (!file->appending)
+          file->reader->set_headers();
 
       set_timestamp_scale();
 
-      for (i = 0; i < g_packetizers.size(); i++)
-        if (g_packetizers[i].packetizer)
-          g_packetizers[i].packetizer->fix_headers();
+      for (auto const &packetizer : g_packetizers)
+        if (packetizer.packetizer)
+          packetizer.packetizer->fix_headers();
 
     } else
       set_timestamp_scale();
@@ -862,7 +932,7 @@ render_attachments(mm_io_c &out) {
 
       std::string name;
       if (attch.stored_name == "")
-        name = bfs::path{attch.name}.filename().string();
+        name = mtx::fs::to_path(attch.name).filename().string();
 
       else
         name = attch.stored_name;
@@ -918,7 +988,7 @@ check_append_mapping() {
     if (!src_file->appending)
       continue;
 
-    size_t count = boost::count_if(g_append_mapping, [&src_file](auto const &e) { return e.src_file_id == src_file->id; });
+    size_t count = std::accumulate(g_append_mapping.begin(), g_append_mapping.end(), 0, [&src_file](auto count, auto const &e) { return count + (e.src_file_id == src_file->id ? 1 : 0); });
 
     if ((0 < count) && (src_file-> reader->m_used_track_ids.size() > count))
       mxerror(fmt::format(Y("Only partial append mappings were given for the file no. {0} ('{1}'). Either don't specify any mapping (in which case the "
@@ -1011,9 +1081,11 @@ check_append_mapping() {
 
     else if (CAN_CONNECT_YES != result) {
       if (error_message.empty())
-        error_message = (  result == CAN_CONNECT_NO_FORMAT     ? Y("The formats do not match.")
-                         : result == CAN_CONNECT_NO_PARAMETERS ? Y("The track parameters do not match.")
-                         :                                       Y("The reason is unknown."));
+        error_message = result == CAN_CONNECT_NO_FORMAT      ? Y("The formats do not match.")
+                      : result == CAN_CONNECT_NO_PARAMETERS  ? Y("The track parameters do not match.")
+                      : result == CAN_CONNECT_NO_UNSUPPORTED ? Y("Appending tracks of this type is not supported.")
+                      :                                        Y("The reason is unknown.");
+
       mxerror(fmt::format(Y("The track number {0} from the file '{1}' cannot be appended to the track number {2} from the file '{3}'. {4}\n"),
                           amap.src_track_id, g_files[amap.src_file_id]->name,
                           amap.dst_track_id, g_files[amap.dst_file_id]->name,
@@ -1028,7 +1100,7 @@ check_append_mapping() {
   // concatenated files. This is needed for displaying the progress.
   for (auto amap = g_append_mapping.begin(), amap_end = g_append_mapping.end(); amap != amap_end; ++amap) {
     // Is this the first in a chain?
-    auto cmp_amap = boost::find_if(g_append_mapping, [&amap](auto const &e) {
+    auto cmp_amap = std::find_if(g_append_mapping.begin(), g_append_mapping.end(), [&amap](auto const &e) {
       return (*amap              != e)
           && (amap->dst_file_id  == e.src_file_id)
           && (amap->dst_track_id == e.src_track_id);
@@ -1039,19 +1111,38 @@ check_append_mapping() {
 
     // Find consecutive mappings.
     auto trav_amap  = amap;
-    int path_length = 2;
     do {
       for (cmp_amap = g_append_mapping.begin(); cmp_amap != amap_end; ++cmp_amap)
         if (   (trav_amap->src_file_id  == cmp_amap->dst_file_id)
             && (trav_amap->src_track_id == cmp_amap->dst_track_id)) {
           trav_amap = cmp_amap;
-          path_length++;
           break;
         }
     } while (cmp_amap != amap_end);
   }
 }
 
+void
+check_split_support() {
+  if (!g_cluster_helper->splitting())
+    return;
+
+  for (auto &ptzr_cont : g_packetizers) {
+    std::string error_message;
+
+    auto &packetizer = *ptzr_cont.packetizer;
+    auto result      = packetizer.can_be_split(error_message);
+
+    if (CAN_SPLIT_YES == result)
+      continue;
+
+    if (error_message.empty())
+      error_message = Y("Splitting tracks of this type is not supported.");
+
+    mxerror(fmt::format(Y("The track ID {0} from the file '{1}' cannot be split. {2}\n"), packetizer.m_ti.m_id, packetizer.m_ti.m_fname, error_message));
+  }
+}
+
 /** \brief Add chapters from the readers and calculate the max size
 
    The reader do not add their chapters to the global chapter pool.
@@ -1083,6 +1174,7 @@ calc_max_chapter_size() {
   // Step 2: Fix the mandatory elements and count the size of all chapters.
   s_max_chapter_size = 0;
   if (g_kax_chapters) {
+    mtx::chapters::unify_legacy_and_bcp47_languages_and_countries(*g_kax_chapters);
     fix_mandatory_elements(g_kax_chapters.get());
     g_kax_chapters->UpdateSize(true);
     s_max_chapter_size += g_kax_chapters->ElementSize();
@@ -1144,6 +1236,37 @@ check_track_id_validity() {
   }
 }
 
+static std::string
+get_first_chapter_name_in_this_file() {
+  if (!s_chapters_in_this_file)
+    return {};
+
+  std::optional<std::pair<int64_t, std::string>> first_chapter;
+
+  for (auto const &chapters_child : *s_chapters_in_this_file) {
+    if (!dynamic_cast<KaxEditionEntry *>(chapters_child))
+      continue;
+
+    for (auto const &edition_child : *static_cast<KaxEditionEntry *>(chapters_child)) {
+      if (!dynamic_cast<KaxChapterAtom *>(edition_child))
+        continue;
+
+      auto &chapter_atom     = *static_cast<KaxChapterAtom *>(edition_child);
+      auto chapter_timestamp = static_cast<int64_t>(FindChildValue<KaxChapterTimeStart>(chapter_atom));
+      auto chapter_display   = FindChild<KaxChapterDisplay>(chapter_atom);
+
+      if (   !chapter_display
+          || (   first_chapter
+              && (chapter_timestamp >= first_chapter->first)))
+        continue;
+
+      first_chapter = std::make_pair(chapter_timestamp, to_utf8(FindChildValue<KaxChapterString>(*chapter_display)));
+    }
+  }
+
+  return first_chapter ? first_chapter->second : ""s;
+}
+
 /** \brief Transform the output filename and insert the current file number
 
    Rules and search order:
@@ -1158,16 +1281,21 @@ create_output_name() {
   // First possibility: %d
   int p    = s.find("%d");
   if (0 <= p) {
-    s.replace(p, 2, to_string(g_file_num));
+    s.replace(p, 2, fmt::to_string(g_file_num));
 
     return s;
   }
 
   // Now search for something like %02d
-  auto converted = boost::regex_replace(s, boost::regex{"%(\\d+)d", boost::regex::perl}, "{0:\\1}");
+  auto converted = to_utf8(Q(s).replace(QRegularExpression{"%(\\d+)d"}, "{0:\\1}"));
   if (converted != s)
     return fmt::format(converted, g_file_num);
 
+  // If chapter names are requested, don't force a numeric suffix. The
+  // chapter name template is replaced when closing the file.
+  if (s.find("%c") != std::string::npos)
+    return s;
+
   std::string buffer = fmt::format("-{0:03}", g_file_num);
 
   // See if we can find a '.'.
@@ -1260,7 +1388,7 @@ prepare_tags_for_rendering() {
   fix_mandatory_elements(s_kax_tags.get());
   sort_ebml_master(s_kax_tags.get());
   if (!s_kax_tags->CheckMandatory())
-    mxerror(fmt::format(Y("Some tag elements are missing (this error should not have occured - another similar error should have occured earlier). {0}\n"), BUGMSG));
+    mxerror(fmt::format(Y("Some tag elements are missing (this error should not have occurred - another similar error should have occurred earlier). {0}\n"), BUGMSG));
 
   s_kax_tags->UpdateSize();
   g_tags_size = s_kax_tags->ElementSize();
@@ -1310,15 +1438,62 @@ create_next_output_file() {
   ++g_file_num;
 }
 
+void
+add_split_points_from_remainig_chapter_numbers() {
+  if (g_splitting_by_chapter_numbers.empty() && !g_splitting_by_all_chapters)
+    return;
+
+  if (!g_kax_chapters)
+    return;
+
+  std::vector<int64_t> chapter_start_timestamps;
+  for (auto element : *g_kax_chapters) {
+    auto edition = dynamic_cast<KaxEditionEntry *>(element);
+    if (!edition)
+      continue;
+
+    for (auto sub_element : *edition) {
+      auto atom = dynamic_cast<KaxChapterAtom *>(sub_element);
+      if (!atom)
+        continue;
+
+      chapter_start_timestamps.push_back(FindChildValue<KaxChapterTimeStart, uint64_t>(atom, 0));
+    }
+  }
+
+  std::sort(chapter_start_timestamps.begin(), chapter_start_timestamps.end());
+
+  std::vector<std::pair<split_point_c, unsigned int>> new_split_points;
+
+  for (auto current_number = 1u; current_number <= chapter_start_timestamps.size(); ++current_number) {
+    auto use_chapter = g_splitting_by_all_chapters || (g_splitting_by_chapter_numbers[current_number] == 1);
+    if (!use_chapter)
+      continue;
+
+    if (g_splitting_by_chapter_numbers[current_number] != 2) {
+      auto split_after = chapter_start_timestamps[current_number - 1];
+      if (split_after)
+        new_split_points.emplace_back(std::make_pair(split_point_c{split_after, split_point_c::timestamp, true}, current_number));
+    }
+
+    g_splitting_by_chapter_numbers[current_number] = 2;
+  }
+
+  for (auto &split_point : new_split_points) {
+    mxdebug_if(s_debug_splitting_chapters, fmt::format("Adding split point for chapter number {} at {}\n", split_point.second, split_point.first.m_point));
+    g_cluster_helper->add_split_point(split_point.first);
+  }
+}
+
 static void
 add_chapters_for_current_part() {
-  auto s_debug = debugging_option_c{"splitting_chapters"};
-
-  mxdebug_if(s_debug, fmt::format("Adding chapters. have_global? {0} splitting? {1}\n", !!g_kax_chapters, g_cluster_helper->splitting()));
+  mxdebug_if(s_debug_splitting_chapters, fmt::format("Adding chapters. have_global? {0} splitting? {1}\n", !!g_kax_chapters, g_cluster_helper->splitting()));
 
   if (!g_cluster_helper->splitting()) {
     s_chapters_in_this_file = clone(g_kax_chapters);
     mtx::chapters::merge_entries(*s_chapters_in_this_file);
+    if (mtx::bcp47::language_c::is_disabled())
+      remove_ietf_language_elements(*s_chapters_in_this_file);
     sort_ebml_master(s_chapters_in_this_file.get());
     return;
   }
@@ -1330,7 +1505,7 @@ add_chapters_for_current_part() {
   auto chapters_here              = clone(g_kax_chapters);
   bool have_chapters_in_timeframe = mtx::chapters::select_in_timeframe(chapters_here.get(), start, end, offset);
 
-  mxdebug_if(s_debug, fmt::format("offset {0} start {1} end {2} have chapters in timeframe? {3} chapters in this file? {4}\n", offset, start, end, have_chapters_in_timeframe, !!s_chapters_in_this_file));
+  mxdebug_if(s_debug_splitting_chapters, fmt::format("offset {0} start {1} end {2} have chapters in timeframe? {3} chapters in this file? {4}\n", offset, start, end, have_chapters_in_timeframe, !!s_chapters_in_this_file));
 
   if (!have_chapters_in_timeframe)
     return;
@@ -1347,7 +1522,7 @@ add_chapters_for_current_part() {
 void
 add_chapter_atom(timestamp_c const &start_timestamp,
                  std::string const &name,
-                 std::string const &language) {
+                 mtx::bcp47::language_c const &language) {
   s_additional_chapter_atoms.emplace_back(start_timestamp, name, language);
 }
 
@@ -1370,12 +1545,23 @@ prepare_additional_chapter_atoms_for_rendering() {
   if (!FindChild<KaxEditionUID>(edition))
     GetChild<KaxEditionUID>(edition).SetValue(create_unique_number(UNIQUE_EDITION_IDS));
 
-  for (auto const &additional_chapter : s_additional_chapter_atoms) {
+  for (auto const &[ch_timestamp, ch_name, ch_language] : s_additional_chapter_atoms) {
     auto atom = cons<KaxChapterAtom>(new KaxChapterUID,       create_unique_number(UNIQUE_CHAPTER_IDS),
-                                     new KaxChapterTimeStart, (std::get<0>(additional_chapter) - offset).to_ns());
-    if (!std::get<1>(additional_chapter).empty())
-      atom->PushElement(*cons<KaxChapterDisplay>(new KaxChapterString,   std::get<1>(additional_chapter),
-                                                 new KaxChapterLanguage, std::get<2>(additional_chapter)));
+                                     new KaxChapterTimeStart, (ch_timestamp - offset).to_ns());
+
+    if (!ch_name.empty()) {
+      auto &display = GetChild<KaxChapterDisplay>(*atom);
+
+      GetChild<KaxChapterString>(display).SetValueUTF8(ch_name);
+
+      if (ch_language.is_valid()) {
+        GetChild<KaxChapterLanguage>(display).SetValue(ch_language.get_closest_iso639_2_alpha_3_code());
+        if (!mtx::bcp47::language_c::is_disabled())
+          GetChild<KaxChapLanguageIETF>(display).SetValue(ch_language.format());
+        else
+          DeleteChildren<KaxChapLanguageIETF>(display);
+      }
+    }
 
     edition.PushElement(*atom);
   }
@@ -1400,8 +1586,8 @@ render_chapters() {
 
   auto replaced = false;
   if (s_kax_chapters_void) {
-    auto with_defaults = !outputting_webm();
-    replaced           = s_kax_chapters_void->ReplaceWith(*s_chapters_in_this_file, *s_out, true, with_defaults);
+    g_doc_type_version_handler->account(*s_chapters_in_this_file);
+    replaced = s_kax_chapters_void->ReplaceWith(*s_chapters_in_this_file, *s_out, true, true);
   }
 
   if (!replaced) {
@@ -1425,6 +1611,36 @@ set_track_statistics_tags(KaxTags *tags) {
   return tags;
 }
 
+static void
+insert_chapter_name_in_output_file_name(boost::filesystem::path const &original_file_name,
+                                        std::string const &chapter_name) {
+#if defined(SYS_WINDOWS)
+  static QRegularExpression s_invalid_char_re{R"([\\/:<>"|?*]+)"};
+#else
+  static QRegularExpression s_invalid_char_re{"/+"};
+#endif
+
+  // When splitting replace %c in file names with current chapter name.
+  if (!g_cluster_helper->split_mode_produces_many_files())
+    return;
+
+  // auto chapter_name  = get_current_chapter_name();
+  auto cleaned_chapter_name = Q(chapter_name).replace(s_invalid_char_re, "-");
+  auto new_file_name        = original_file_name.parent_path() / mtx::fs::to_path(Q(original_file_name.filename()).replace(QRegularExpression{"%c"}, cleaned_chapter_name));
+
+  mxdebug_if(s_debug_splitting_chapters, fmt::format("insert_chapter_name_in_output_file_name: cleaned name {0} old {1} new {2}\n", to_utf8(cleaned_chapter_name), original_file_name.string(), new_file_name.string()));
+
+  if (original_file_name == new_file_name)
+    return;
+
+  try {
+    boost::filesystem::rename(original_file_name, new_file_name);
+    mxinfo(fmt::format(Y("The file '{0}' was renamed to '{1}'.\n"), original_file_name.string(), new_file_name.string()));
+  } catch (boost::filesystem::filesystem_error &) {
+    mxerror(fmt::format(Y("The file '{0}' could not be renamed to '{1}'.\n"), original_file_name.string(), new_file_name.string()));
+  }
+}
+
 /** \brief Finishes and closes the current file
 
    Renders the data that is generated during the muxing run. The cues
@@ -1446,8 +1662,6 @@ finish_file(bool last_file,
   run_before_file_finished_packetizer_hooks();
 
   bool do_output = verbose && !dynamic_cast<mm_null_io_c *>(s_out.get());
-  if (do_output)
-    mxinfo("\n");
 
   // Render the track headers a second time if the user has requested that.
   if (mtx::hacks::is_engaged(mtx::hacks::WRITE_HEADERS_TWICE)) {
@@ -1549,6 +1763,9 @@ finish_file(bool last_file,
 
   if (tags_here) {
     fix_mandatory_elements(tags_here);
+    if (mtx::bcp47::language_c::is_disabled())
+      remove_ietf_language_elements(*tags_here);
+    remove_mandatory_elements_set_to_their_default(*tags_here);
     tags_here->UpdateSize();
     g_doc_type_version_handler->render(*tags_here, *s_out, true);
 
@@ -1556,6 +1773,8 @@ finish_file(bool last_file,
     delete tags_here;
   }
 
+  auto first_chapter_name = get_first_chapter_name_in_this_file();
+
   if (s_chapters_in_this_file) {
     if (!mtx::hacks::is_engaged(mtx::hacks::NO_CHAPTERS_IN_META_SEEK))
       g_kax_sh_main->IndexThis(*s_chapters_in_this_file, *g_kax_segment);
@@ -1581,6 +1800,8 @@ finish_file(bool last_file,
 
   update_ebml_head();
 
+  auto original_file_name = mtx::fs::to_path(s_out->get_file_name());
+
   s_out.reset();
 
   g_kax_segment.reset();
@@ -1590,6 +1811,8 @@ finish_file(bool last_file,
   g_kax_sh_cues.reset();
   s_head.reset();
   g_doc_type_version_handler.reset();
+
+  insert_chapter_name_in_output_file_name(original_file_name, first_chapter_name);
 }
 
 void
@@ -1648,7 +1871,7 @@ append_track(packetizer_t &ptzr,
 
   // Find the generic_packetizer_c that we will be appending to the one
   // stored in ptzr.
-  auto gptzr = brng::find_if(src_file.reader->m_reader_packetizers, [&amap](generic_packetizer_c *p) -> bool {
+  auto gptzr = std::find_if(src_file.reader->m_reader_packetizers.begin(), src_file.reader->m_reader_packetizers.end(), [&amap](auto p) -> bool {
     return amap.src_track_id == static_cast<size_t>(p->m_ti.m_id);
   });
 
@@ -1679,7 +1902,7 @@ append_track(packetizer_t &ptzr,
       if (file->done)
         continue;
 
-      auto vptzr = brng::find_if(file->reader->m_reader_packetizers, [](generic_packetizer_c *p) { return p->get_track_type() == track_video; });
+      auto vptzr = std::find_if(file->reader->m_reader_packetizers.begin(), file->reader->m_reader_packetizers.end(), [](auto p) { return p->get_track_type() == track_video; });
       if (vptzr == file->reader->m_reader_packetizers.end())
         continue;
 
@@ -1696,9 +1919,9 @@ append_track(packetizer_t &ptzr,
 
   if (s_debug_appending) {
     mxdebug(fmt::format("appending: reader m_max_timestamp_seen {0} and ptzr to append is {1} ptzr appended to is {2} src_file.appending {3} src_file.appended_to {4} dst_file.appending {5} dst_file.appended_to {6}\n",
-                        dst_file.reader->m_max_timestamp_seen, static_cast<void *>(*gptzr), static_cast<void *>(ptzr.packetizer), src_file.appending, src_file.appended_to, dst_file.appending, dst_file.appended_to));
+                        dst_file.reader->m_max_timestamp_seen, static_cast<void *>(gptzr->get()), static_cast<void *>(ptzr.packetizer), src_file.appending, src_file.appended_to, dst_file.appending, dst_file.appended_to));
     for (auto &rep_ptzr : dst_file.reader->m_reader_packetizers)
-      mxdebug(fmt::format("  ptzr @ {0} connected_to {1} max_timestamp_seen {2}\n", static_cast<void *>(rep_ptzr), rep_ptzr->m_connected_to, rep_ptzr->m_max_timestamp_seen));
+      mxdebug(fmt::format("  ptzr @ {0} connected_to {1} max_timestamp_seen {2}\n", static_cast<void *>(rep_ptzr.get()), rep_ptzr->m_connected_to, rep_ptzr->m_max_timestamp_seen));
   }
 
   // In rare cases (e.g. empty tracks) a whole file could be skipped
@@ -1712,8 +1935,8 @@ append_track(packetizer_t &ptzr,
   if (dst_file.appending) {
     std::list<generic_packetizer_c *> not_connected_ptzrs;
     for (auto &check_ptzr : dst_file.reader->m_reader_packetizers)
-      if ((check_ptzr != ptzr.packetizer) && (2 != check_ptzr->m_connected_to))
-        not_connected_ptzrs.push_back(check_ptzr);
+      if ((check_ptzr.get() != ptzr.packetizer) && (2 != check_ptzr->m_connected_to))
+        not_connected_ptzrs.push_back(check_ptzr.get());
 
     if (s_debug_appending) {
       std::string result;
@@ -1732,7 +1955,7 @@ append_track(packetizer_t &ptzr,
 
   // Also fix the ptzr structure and reset the ptzr's state to "I want more".
   generic_packetizer_c *old_packetizer = ptzr.packetizer;
-  ptzr.packetizer                      = *gptzr;
+  ptzr.packetizer                      = (*gptzr).get();
   ptzr.file                            = amap.src_file_id;
   ptzr.status                          = FILE_STATUS_MOREDATA;
 
@@ -1776,7 +1999,7 @@ append_track(packetizer_t &ptzr,
       ptzr.status = ptzr.packetizer->read(false);
 
     if (src_file.reader->m_ptzr_first_packet) {
-      auto cmp_amap = brng::find_if(g_append_mapping, [&amap, &src_file](append_spec_t const &m) {
+      auto cmp_amap = std::find_if(g_append_mapping.begin(), g_append_mapping.end(), [&amap, &src_file](append_spec_t const &m) {
         return (m.src_file_id  == amap.src_file_id)
             && (m.src_track_id == static_cast<size_t>(src_file.reader->m_ptzr_first_packet->m_ti.m_id))
             && (m.dst_file_id  == amap.dst_file_id);
@@ -1791,12 +2014,12 @@ append_track(packetizer_t &ptzr,
   }
 
   if ((APPEND_MODE_FILE_BASED == g_append_mode) || (ptzr.packetizer->get_track_type() == track_subtitle)) {
-    mxdebug_if(s_debug_appending, fmt::format("appending: new timestamp_adjustment for append_mode == FILE_BASED or subtitle track: {0} for {1}\n", format_timestamp(timestamp_adjustment), ptzr.packetizer->m_ti.m_id));
+    mxdebug_if(s_debug_appending, fmt::format("appending: new timestamp_adjustment for append_mode == FILE_BASED or subtitle track: {0} for {1}\n", mtx::string::format_timestamp(timestamp_adjustment), ptzr.packetizer->m_ti.m_id));
     // The actual connection.
     ptzr.packetizer->connect(old_packetizer, timestamp_adjustment);
 
   } else {
-    mxdebug_if(s_debug_appending, fmt::format("appending: new timestamp_adjustment for append_mode == TRACK_BASED and NON subtitle track: {0} for {1}\n", format_timestamp(timestamp_adjustment), ptzr.packetizer->m_ti.m_id));
+    mxdebug_if(s_debug_appending, fmt::format("appending: new timestamp_adjustment for append_mode == TRACK_BASED and NON subtitle track: {0} for {1}\n", mtx::string::format_timestamp(timestamp_adjustment), ptzr.packetizer->m_ti.m_id));
     // The actual connection.
     ptzr.packetizer->connect(old_packetizer);
   }
@@ -1870,8 +2093,16 @@ establish_deferred_connections(filelist_t &file) {
   // \todo Select a new file that the subs will defer to.
 }
 
-static void
+static bool
 check_and_handle_end_of_input_after_pulling(packetizer_t &ptzr) {
+  auto end_of_video_reached = false;
+
+  if (   !ptzr.pack
+      && g_video_packetizer
+      && (g_video_packetizer == ptzr.packetizer)
+      && mtx::included_in(ptzr.status, FILE_STATUS_DONE, FILE_STATUS_HOLDING))
+    end_of_video_reached = true;
+
   if (!ptzr.pack && (FILE_STATUS_DONE == ptzr.status))
     ptzr.status = FILE_STATUS_DONE_AND_DRY;
 
@@ -1880,7 +2111,7 @@ check_and_handle_end_of_input_after_pulling(packetizer_t &ptzr) {
   // unfinished packetizers in the corresponding file structure.
   if (   (FILE_STATUS_DONE_AND_DRY != ptzr.status)
       || (ptzr.old_status          == ptzr.status))
-    return;
+    return end_of_video_reached;
 
   auto &file = *g_files[ptzr.file];
   file.num_unfinished_packetizers--;
@@ -1892,11 +2123,14 @@ check_and_handle_end_of_input_after_pulling(packetizer_t &ptzr) {
     file.done = true;
   }
   file.old_num_unfinished_packetizers = file.num_unfinished_packetizers;
+
+  return end_of_video_reached;
 }
 
-static bool
+static std::pair<bool, bool>
 force_pull_packetizers_of_fully_held_files() {
   std::unordered_map<generic_reader_c *, bool> fully_held_files;
+  auto end_of_video_reached = false;
 
   for (auto &ptzr : g_packetizers) {
     auto reader = ptzr.packetizer->m_reader;
@@ -1919,14 +2153,17 @@ force_pull_packetizers_of_fully_held_files() {
       if (!ptzr.pack)
         ptzr.pack = ptzr.packetizer->get_packet();
 
-      check_and_handle_end_of_input_after_pulling(ptzr);
+      if (check_and_handle_end_of_input_after_pulling(ptzr))
+        end_of_video_reached = true;
     }
 
-  return force_pulled;
+  return { end_of_video_reached, force_pulled };
 }
 
-static void
+static bool
 pull_packetizers_for_packets() {
+  auto end_of_video_reached = false;
+
   for (auto &ptzr : g_packetizers) {
     if (FILE_STATUS_HOLDING == ptzr.status)
       ptzr.status = FILE_STATUS_MOREDATA;
@@ -1945,8 +2182,11 @@ pull_packetizers_for_packets() {
     if (!ptzr.pack)
       ptzr.pack = ptzr.packetizer->get_packet();
 
-    check_and_handle_end_of_input_after_pulling(ptzr);
+    if (check_and_handle_end_of_input_after_pulling(ptzr))
+      end_of_video_reached = true;
   }
+
+  return end_of_video_reached;
 }
 
 static packetizer_t *
@@ -1987,8 +2227,11 @@ main_loop() {
   while (1) {
     // Step 1: Make sure a packet is available for each output
     // as long we haven't already processed the last one.
-    pull_packetizers_for_packets();
-    auto force_pulled = force_pull_packetizers_of_fully_held_files();
+    auto end_of_video_reached1                 = pull_packetizers_for_packets();
+    auto [end_of_video_reached2, force_pulled] = force_pull_packetizers_of_fully_held_files();
+
+    if (g_stop_after_video_ends && (end_of_video_reached1 || end_of_video_reached2))
+      break;
 
     // Step 2: Pick the packet with the lowest timestamp and
     // stuff it into the Matroska file.
@@ -1998,14 +2241,14 @@ main_loop() {
     bool appended_a_track = s_appending_files && append_tracks_maybe();
 
     if (winner && winner->pack) {
-      packet_cptr pack = winner->pack;
-
       // Step 3: Add the winning packet to a cluster. Full clusters will be
       // rendered automatically.
-      g_cluster_helper->add_packet(pack);
+      g_cluster_helper->add_packet(winner->pack);
 
       winner->pack.reset();
 
+      add_split_points_from_remainig_chapter_numbers();
+
       // If splitting by parts is active and the last part has been
       // processed fully then we can finish up.
       if (g_cluster_helper->is_splitting_and_processed_fully()) {
