@@ -24,6 +24,7 @@
 #include "common/segmentinfo.h"
 #include "common/strings/formatting.h"
 #include "common/unique_numbers.h"
+#include "mkvtoolnix-gui/app.h"
 #include "mkvtoolnix-gui/forms/header_editor/tab.h"
 #include "mkvtoolnix-gui/header_editor/action_for_dropped_files_dialog.h"
 #include "mkvtoolnix-gui/header_editor/ascii_string_value_page.h"
@@ -66,6 +67,8 @@ Tab::Tab(QWidget *parent,
   , m_fileName{fileName}
   , m_model{new PageModel{this}}
   , m_treeContextMenu{new QMenu{this}}
+  , m_modifySelectedTrackMenu{new QMenu{this}}
+  , m_languageShortcutsMenu{new QMenu{this}}
   , m_expandAllAction{new QAction{this}}
   , m_collapseAllAction{new QAction{this}}
   , m_addAttachmentsAction{new QAction{this}}
@@ -93,6 +96,7 @@ Tab::resetData() {
   m_eTracks.reset();
   m_model->reset();
   m_segmentinfoPage = nullptr;
+  m_tracksReordered = false;
 }
 
 void
@@ -129,7 +133,7 @@ Tab::load() {
     return;
   }
 
-  m_analyzer = std::make_unique<QtKaxAnalyzer>(this, m_fileName);
+  m_analyzer = std::make_unique<Util::KaxAnalyzer>(this, m_fileName);
   bool ok    = false;
   QString error;
 
@@ -196,7 +200,7 @@ Tab::save() {
       tracksModified      = true;
   }
 
-  if (!segmentinfoModified && !tracksModified && !attachmentsModified) {
+  if (!segmentinfoModified && !tracksModified && !attachmentsModified && !m_tracksReordered) {
     Util::MessageBox::information(this)->title(QY("File has not been modified")).text(QY("The header values have not been modified. There is nothing to save.")).exec();
     return;
   }
@@ -221,15 +225,17 @@ Tab::save() {
     if (segmentinfoModified && m_eSegmentInfo) {
       auto result = m_analyzer->update_element(m_eSegmentInfo, true);
       if (kax_analyzer_c::uer_success != result) {
-        QtKaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified segment information header failed."));
+        Util::KaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified segment information header failed."));
         ok = false;
       }
     }
 
-    if (ok && tracksModified && m_eTracks) {
+    if (ok && m_eTracks && (tracksModified || m_tracksReordered)) {
+      updateTracksElementToMatchTrackOrder();
+
       auto result = m_analyzer->update_element(m_eTracks, true);
       if (kax_analyzer_c::uer_success != result) {
-        QtKaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified track headers failed."));
+        Util::KaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified track headers failed."));
         ok = false;
       }
     }
@@ -241,12 +247,12 @@ Tab::save() {
         attachments->PushElement(*dynamic_cast<AttachedFilePage &>(*attachedFilePage).m_attachment.get());
 
       auto result = attachments->ListSize() ? m_analyzer->update_element(attachments.get(), true)
-                  :                           m_analyzer->remove_elements(KaxAttachments::ClassInfos.GlobalId);
+                  :                           m_analyzer->remove_elements(EBML_ID(KaxAttachments));
 
       attachments->RemoveAll();
 
       if (kax_analyzer_c::uer_success != result) {
-        QtKaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified attachments failed."));
+        Util::KaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified attachments failed."));
         ok = false;
       }
     }
@@ -255,7 +261,7 @@ Tab::save() {
       auto result = m_analyzer->update_uid_referrals(trackUIDChanges);
 
       if (kax_analyzer_c::uer_success != result) {
-        QtKaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified attachments failed."));
+        Util::KaxAnalyzer::displayUpdateElementResult(this, result, QY("Saving the modified attachments failed."));
         ok = false;
       }
     }
@@ -288,6 +294,8 @@ Tab::save() {
 
 void
 Tab::setupUi() {
+  setupModifyTracksMenu();
+
   Util::Settings::get().handleSplitterSizes(ui->headerEditorSplitter);
 
   auto info = QFileInfo{m_fileName};
@@ -300,20 +308,50 @@ Tab::setupUi() {
   Util::HeaderViewManager::create(*ui->elements, "HeaderEditor::Elements").setDefaultSizes({ { Q("type"), 250 }, { Q("codec"), 100 }, { Q("language"), 120 }, { Q("properties"), 120 } });
   Util::preventScrollingWithoutFocus(this);
 
-  connect(ui->elements,                              &Util::BasicTreeView::customContextMenuRequested, this, &Tab::showTreeContextMenu);
-  connect(ui->elements,                              &Util::BasicTreeView::filesDropped,               this, &Tab::handleDroppedFiles);
-  connect(ui->elements,                              &Util::BasicTreeView::deletePressed,              this, &Tab::removeSelectedAttachment);
-  connect(ui->elements,                              &Util::BasicTreeView::insertPressed,              this, &Tab::selectAttachmentsAndAdd);
-  connect(ui->elements->selectionModel(),            &QItemSelectionModel::currentChanged,             this, &Tab::selectionChanged);
-  connect(m_expandAllAction,                         &QAction::triggered,                              this, &Tab::expandAll);
-  connect(m_collapseAllAction,                       &QAction::triggered,                              this, &Tab::collapseAll);
-  connect(m_addAttachmentsAction,                    &QAction::triggered,                              this, &Tab::selectAttachmentsAndAdd);
-  connect(m_removeAttachmentAction,                  &QAction::triggered,                              this, &Tab::removeSelectedAttachment);
-  connect(m_removeAllAttachmentsAction,              &QAction::triggered,                              this, &Tab::removeAllAttachments);
-  connect(m_saveAttachmentContentAction,             &QAction::triggered,                              this, &Tab::saveAttachmentContent);
-  connect(m_replaceAttachmentContentAction,          &QAction::triggered,                              [this]() { replaceAttachmentContent(false); });
-  connect(m_replaceAttachmentContentSetValuesAction, &QAction::triggered,                              [this]() { replaceAttachmentContent(true); });
-  connect(m_model,                                   &PageModel::attachmentsReordered,                 [this]() { m_attachmentsPage->rereadChildren(*m_model); });
+  auto &mts = m_modifyTracksSubmenu;
+
+  connect(ui->elements,                              &Util::BasicTreeView::customContextMenuRequested,    this, &Tab::showTreeContextMenu);
+  connect(ui->elements,                              &Util::BasicTreeView::filesDropped,                  this, &Tab::handleDroppedFiles);
+  connect(ui->elements,                              &Util::BasicTreeView::deletePressed,                 this, &Tab::removeSelectedAttachment);
+  connect(ui->elements,                              &Util::BasicTreeView::insertPressed,                 this, &Tab::selectAttachmentsAndAdd);
+  connect(ui->elements,                              &Util::BasicTreeView::ctrlDownPressed,               this, [this]() { moveElementUpOrDown(false); });
+  connect(ui->elements,                              &Util::BasicTreeView::ctrlUpPressed,                 this, [this]() { moveElementUpOrDown(true); });
+  connect(ui->elements->selectionModel(),            &QItemSelectionModel::currentChanged,                this, &Tab::selectionChanged);
+  connect(m_expandAllAction,                         &QAction::triggered,                                 this, &Tab::expandAll);
+  connect(m_collapseAllAction,                       &QAction::triggered,                                 this, &Tab::collapseAll);
+  connect(m_addAttachmentsAction,                    &QAction::triggered,                                 this, &Tab::selectAttachmentsAndAdd);
+  connect(m_removeAttachmentAction,                  &QAction::triggered,                                 this, &Tab::removeSelectedAttachment);
+  connect(m_removeAllAttachmentsAction,              &QAction::triggered,                                 this, &Tab::removeAllAttachments);
+  connect(m_saveAttachmentContentAction,             &QAction::triggered,                                 this, &Tab::saveAttachmentContent);
+  connect(mts.m_toggleTrackEnabledFlag,              &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleDefaultTrackFlag,              &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleForcedDisplayFlag,             &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleCommentaryFlag,                &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleOriginalFlag,                  &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleHearingImpairedFlag,           &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleVisualImpairedFlag,            &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(mts.m_toggleTextDescriptionsFlag,          &QAction::triggered,                                 this, &Tab::toggleTrackFlag);
+  connect(&mts,                                      &Util::ModifyTracksSubmenu::languageChangeRequested, this, &Tab::changeTrackLanguage);
+  connect(m_replaceAttachmentContentAction,          &QAction::triggered,                                 [this]() { replaceAttachmentContent(false); });
+  connect(m_replaceAttachmentContentSetValuesAction, &QAction::triggered,                                 [this]() { replaceAttachmentContent(true); });
+  connect(m_model,                                   &PageModel::attachmentsReordered,                    [this]() { m_attachmentsPage->rereadChildren(*m_model); });
+  connect(m_model,                                   &PageModel::tracksReordered,                         this, &Tab::handleReorderedTracks);
+  connect(MainWindow::get(),                         &MainWindow::preferencesChanged,                     [this]() { m_modifyTracksSubmenu.setupLanguage(*m_languageShortcutsMenu); });
+}
+
+void
+Tab::setupModifyTracksMenu() {
+  m_modifySelectedTrackMenu->addMenu(m_languageShortcutsMenu);
+  m_modifySelectedTrackMenu->addSeparator();
+
+  m_modifyTracksSubmenu.setupTrack(*m_modifySelectedTrackMenu);
+  m_modifyTracksSubmenu.setupLanguage(*m_languageShortcutsMenu);
+}
+
+void
+Tab::handleReorderedTracks() {
+  m_tracksReordered = true;
+  m_model->rereadTopLevelPageIndexes();
 }
 
 void
@@ -348,11 +386,15 @@ Tab::retranslateUi() {
   m_saveAttachmentContentAction->setText(QY("&Save attachment content to a file"));
   m_replaceAttachmentContentAction->setText(QY("Re&place attachment with a new file"));
   m_replaceAttachmentContentSetValuesAction->setText(QY("Replace attachment with a new file and &derive name && MIME type from it"));
+  m_modifySelectedTrackMenu->setTitle(QY("Modif&y selected track"));
 
-  m_addAttachmentsAction->setIcon(QIcon{Q(":/icons/16x16/list-add.png")});
-  m_removeAttachmentAction->setIcon(QIcon{Q(":/icons/16x16/list-remove.png")});
-  m_saveAttachmentContentAction->setIcon(QIcon{Q(":/icons/16x16/document-save.png")});
-  m_replaceAttachmentContentAction->setIcon(QIcon{Q(":/icons/16x16/document-open.png")});
+  m_addAttachmentsAction->setIcon(QIcon::fromTheme(Q("list-add")));
+  m_removeAttachmentAction->setIcon(QIcon::fromTheme(Q("list-remove")));
+  m_saveAttachmentContentAction->setIcon(QIcon::fromTheme(Q("document-save")));
+  m_replaceAttachmentContentAction->setIcon(QIcon::fromTheme(Q("document-open")));
+
+  m_modifyTracksSubmenu.retranslateUi();
+  m_languageShortcutsMenu->setTitle(QY("Set &language"));
 
   setupToolTips();
 
@@ -369,15 +411,19 @@ Tab::setupToolTips() {
 
 void
 Tab::populateTree() {
-  m_analyzer->with_elements(KaxInfo::ClassInfos.GlobalId, [this](kax_analyzer_data_c const &data) {
+  m_analyzer->with_elements(EBML_ID(KaxInfo), [this](kax_analyzer_data_c const &data) {
     handleSegmentInfo(data);
   });
 
-  m_analyzer->with_elements(KaxTracks::ClassInfos.GlobalId, [this](kax_analyzer_data_c const &data) {
+  m_analyzer->with_elements(EBML_ID(KaxTracks), [this](kax_analyzer_data_c const &data) {
     handleTracks(data);
   });
 
   handleAttachments();
+
+  for (auto const &page : m_model->topLevelPages())
+    if (dynamic_cast<TrackTypePage *>(page))
+      static_cast<TrackTypePage *>(page)->updateModelItems();
 }
 
 void
@@ -393,6 +439,16 @@ Tab::selectionChanged(QModelIndex const &current,
     ui->pageContainer->setCurrentWidget(selectedPage);
 }
 
+bool
+Tab::isTrackSelected() {
+  auto topLevelIdx = ui->elements->selectionModel()->currentIndex();
+
+  while (topLevelIdx.parent().isValid())
+    topLevelIdx = topLevelIdx.parent();
+
+  return !!dynamic_cast<TrackTypePage *>(m_model->selectedPage(topLevelIdx));
+}
+
 QString const &
 Tab::fileName()
   const {
@@ -407,8 +463,7 @@ Tab::title()
 
 PageBase *
 Tab::hasBeenModified() {
-  auto &pages = m_model->topLevelPages();
-  for (auto const &page : pages) {
+  for (auto const &page : m_model->topLevelPages()) {
     auto modifiedPage = page->hasBeenModified();
     if (modifiedPage)
       return modifiedPage;
@@ -427,15 +482,15 @@ Tab::pruneEmptyMastersForTrack(TrackTypePage &page) {
   std::unordered_map<EbmlMaster *, bool> handled;
 
   if (trackType == track_video) {
-    auto trackVideo            = &GetChildEmptyIfNew<KaxTrackVideo>(page.m_master);
-    auto videoColour           = &GetChildEmptyIfNew<KaxVideoColour>(trackVideo);
-    auto videoColourMasterMeta = &GetChildEmptyIfNew<KaxVideoColourMasterMeta>(videoColour);
-    auto videoProjection       = &GetChildEmptyIfNew<KaxVideoProjection>(trackVideo);
+    auto trackVideo           = &GetChildEmptyIfNew<KaxTrackVideo>(page.m_master);
+    auto videoColor           = &GetChildEmptyIfNew<KaxVideoColour>(trackVideo);
+    auto videoColorMasterMeta = &GetChildEmptyIfNew<KaxVideoColourMasterMeta>(videoColor);
+    auto videoProjection      = &GetChildEmptyIfNew<KaxVideoProjection>(trackVideo);
 
-    remove_master_from_parent_if_empty_or_only_defaults(videoColour,    videoColourMasterMeta, handled);
-    remove_master_from_parent_if_empty_or_only_defaults(trackVideo,     videoColour,           handled);
-    remove_master_from_parent_if_empty_or_only_defaults(trackVideo,     videoProjection,       handled);
-    remove_master_from_parent_if_empty_or_only_defaults(&page.m_master, trackVideo,            handled);
+    remove_master_from_parent_if_empty_or_only_defaults(videoColor,     videoColorMasterMeta, handled);
+    remove_master_from_parent_if_empty_or_only_defaults(trackVideo,     videoColor,           handled);
+    remove_master_from_parent_if_empty_or_only_defaults(trackVideo,     videoProjection,      handled);
+    remove_master_from_parent_if_empty_or_only_defaults(&page.m_master, trackVideo,           handled);
 
   } else
     // trackType is track_audio
@@ -459,7 +514,7 @@ Tab::determineTrackUIDChanges() {
       if (!uiValuePage)
         continue;
 
-      if (uiValuePage->m_callbacks.GlobalId != KaxTrackUID::ClassInfos.GlobalId)
+      if (uiValuePage->m_callbacks.ClassId() != EBML_ID(KaxTrackUID))
         continue;
 
       if (uiValuePage->m_cbAddOrRemove->isChecked())
@@ -476,8 +531,7 @@ Tab::determineTrackUIDChanges() {
 
 void
 Tab::doModifications() {
-  auto &pages = m_model->topLevelPages();
-  for (auto const &page : pages)
+  for (auto const &page : m_model->topLevelPages())
     page->doModifications();
 
   pruneEmptyMastersForAllTracks();
@@ -500,9 +554,9 @@ Tab::createValuePage(TopLevelPage &parentPage,
   ValuePage *page{};
   auto const type = element.m_type;
 
-  page = element.m_callbacks == &KaxTrackLanguage::ClassInfos     ? new LanguageValuePage{       *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
-       : element.m_callbacks == &KaxLanguageIETF::ClassInfos      ? new LanguageIETFValuePage{   *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
-       : element.m_callbacks == &KaxTrackName::ClassInfos         ? new TrackNamePage{           *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
+  page = element.m_callbacks == &EBML_INFO(KaxTrackLanguage)      ? new LanguageValuePage{       *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
+       : element.m_callbacks == &EBML_INFO(KaxLanguageIETF)       ? new LanguageIETFValuePage{   *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
+       : element.m_callbacks == &EBML_INFO(KaxTrackName)          ? new TrackNamePage{           *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
        : type                == property_element_c::EBMLT_BOOL    ? new BoolValuePage{           *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
        : type                == property_element_c::EBMLT_BINARY  ? new BitValuePage{            *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description, element.m_bit_length}
        : type                == property_element_c::EBMLT_FLOAT   ? new FloatValuePage{          *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
@@ -511,10 +565,12 @@ Tab::createValuePage(TopLevelPage &parentPage,
        : type                == property_element_c::EBMLT_STRING  ? new AsciiStringValuePage{    *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
        : type                == property_element_c::EBMLT_USTRING ? new StringValuePage{         *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
        : type                == property_element_c::EBMLT_DATE    ? new TimeValuePage{           *this, parentPage, parentMaster, *element.m_callbacks, element.m_title, element.m_description}
-       :                                                             static_cast<ValuePage *>(nullptr);
+       :                                                            static_cast<ValuePage *>(nullptr);
 
-  if (page)
+  if (page) {
     page->init();
+    connect(page, &ValuePage::valueChanged, static_cast<TrackTypePage *>(&parentPage), &TrackTypePage::updateModelItems);
+  }
 
   return page;
 }
@@ -530,7 +586,7 @@ Tab::handleSegmentInfo(kax_analyzer_data_c const &data) {
   page->setInternalIdentifier("segmentInfo");
   page->init();
 
-  auto &propertyElements = property_element_c::get_table_for(KaxInfo::ClassInfos, nullptr, true);
+  auto &propertyElements = property_element_c::get_table_for(EBML_INFO(KaxInfo), nullptr, true);
   for (auto const &element : propertyElements)
     createValuePage(*page, info, element);
 
@@ -544,7 +600,7 @@ Tab::handleTracks(kax_analyzer_data_c const &data) {
     return;
 
   auto trackIdxMkvmerge  = 0u;
-  auto &propertyElements = property_element_c::get_table_for(KaxTracks::ClassInfos, nullptr, true);
+  auto &propertyElements = property_element_c::get_table_for(EBML_INFO(KaxTracks), nullptr, true);
 
   for (auto const &element : dynamic_cast<EbmlMaster &>(*m_eTracks)) {
     auto kTrackEntry = dynamic_cast<KaxTrackEntry *>(element);
@@ -566,34 +622,34 @@ Tab::handleTracks(kax_analyzer_data_c const &data) {
     parentPagesByCallback[nullptr]   = page;
 
     if (track_video == trackType) {
-      auto colourPage = new TopLevelPage{*this, YT("Colour information")};
-      colourPage->setInternalIdentifier(Q("videoColour %1").arg(trackIdxMkvmerge - 1));
-      colourPage->setParentPage(*page);
-      colourPage->init();
+      auto colorPage = new TopLevelPage{*this, YT("Color information")};
+      colorPage->setInternalIdentifier(Q("videoColor %1").arg(trackIdxMkvmerge - 1));
+      colorPage->setParentPage(*page);
+      colorPage->init();
 
-      auto colourMasterMetaPage = new TopLevelPage{*this, YT("Colour mastering meta information")};
-      colourMasterMetaPage->setInternalIdentifier(Q("videoColourMasterMeta %1").arg(trackIdxMkvmerge - 1));
-      colourMasterMetaPage->setParentPage(*page);
-      colourMasterMetaPage->init();
+      auto colorMasterMetaPage = new TopLevelPage{*this, YT("Color mastering meta information")};
+      colorMasterMetaPage->setInternalIdentifier(Q("videoColorMasterMeta %1").arg(trackIdxMkvmerge - 1));
+      colorMasterMetaPage->setParentPage(*page);
+      colorMasterMetaPage->init();
 
       auto projectionPage = new TopLevelPage{*this, YT("Video projection information")};
       projectionPage->setInternalIdentifier(Q("videoProjection %1").arg(trackIdxMkvmerge - 1));
       projectionPage->setParentPage(*page);
       projectionPage->init();
 
-      parentMastersByCallback[&KaxTrackVideo::ClassInfos]            = &GetChildEmptyIfNew<KaxTrackVideo>(kTrackEntry);
-      parentMastersByCallback[&KaxVideoColour::ClassInfos]           = &GetChildEmptyIfNew<KaxVideoColour>(parentMastersByCallback[&KaxTrackVideo::ClassInfos]);
-      parentMastersByCallback[&KaxVideoColourMasterMeta::ClassInfos] = &GetChildEmptyIfNew<KaxVideoColourMasterMeta>(parentMastersByCallback[&KaxVideoColour::ClassInfos]);
-      parentMastersByCallback[&KaxVideoProjection::ClassInfos]       = &GetChildEmptyIfNew<KaxVideoProjection>(parentMastersByCallback[&KaxTrackVideo::ClassInfos]);
+      parentMastersByCallback[&EBML_INFO(KaxTrackVideo)]            = &GetChildEmptyIfNew<KaxTrackVideo>(kTrackEntry);
+      parentMastersByCallback[&EBML_INFO(KaxVideoColour)]           = &GetChildEmptyIfNew<KaxVideoColour>(parentMastersByCallback[&EBML_INFO(KaxTrackVideo)]);
+      parentMastersByCallback[&EBML_INFO(KaxVideoColourMasterMeta)] = &GetChildEmptyIfNew<KaxVideoColourMasterMeta>(parentMastersByCallback[&EBML_INFO(KaxVideoColour)]);
+      parentMastersByCallback[&EBML_INFO(KaxVideoProjection)]       = &GetChildEmptyIfNew<KaxVideoProjection>(parentMastersByCallback[&EBML_INFO(KaxTrackVideo)]);
 
-      parentPagesByCallback[&KaxTrackVideo::ClassInfos]              = page;
-      parentPagesByCallback[&KaxVideoColour::ClassInfos]             = colourPage;
-      parentPagesByCallback[&KaxVideoColourMasterMeta::ClassInfos]   = colourMasterMetaPage;
-      parentPagesByCallback[&KaxVideoProjection::ClassInfos]         = projectionPage;
+      parentPagesByCallback[&EBML_INFO(KaxTrackVideo)]              = page;
+      parentPagesByCallback[&EBML_INFO(KaxVideoColour)]             = colorPage;
+      parentPagesByCallback[&EBML_INFO(KaxVideoColourMasterMeta)]   = colorMasterMetaPage;
+      parentPagesByCallback[&EBML_INFO(KaxVideoProjection)]         = projectionPage;
 
     } else if (track_audio == trackType) {
-      parentMastersByCallback[&KaxTrackAudio::ClassInfos]            = &GetChildEmptyIfNew<KaxTrackAudio>(kTrackEntry);
-      parentPagesByCallback[&KaxTrackAudio::ClassInfos]              = page;
+      parentMastersByCallback[&EBML_INFO(KaxTrackAudio)]            = &GetChildEmptyIfNew<KaxTrackAudio>(kTrackEntry);
+      parentPagesByCallback[&EBML_INFO(KaxTrackAudio)]              = page;
     }
 
     for (auto const &propElement : propertyElements) {
@@ -613,7 +669,7 @@ void
 Tab::handleAttachments() {
   auto attachments = KaxAttachedList{};
 
-  m_analyzer->with_elements(KaxAttachments::ClassInfos.GlobalId, [this, &attachments](kax_analyzer_data_c const &data) {
+  m_analyzer->with_elements(EBML_ID(KaxAttachments), [this, &attachments](kax_analyzer_data_c const &data) {
     auto master = std::dynamic_pointer_cast<KaxAttachments>(m_analyzer->read_element(data));
     if (!master)
       return;
@@ -675,6 +731,7 @@ Tab::showTreeContextMenu(QPoint const &pos) {
   auto isAttachmentsPage  = !!dynamic_cast<AttachmentsPage *>(selectedPage);
   auto isAttachedFilePage = !!dynamic_cast<AttachedFilePage *>(selectedPage);
   auto isAttachments      = isAttachmentsPage || isAttachedFilePage;
+  auto isTrack            = isTrackSelected();
   auto actions            = m_treeContextMenu->actions();
 
   for (auto const &action : actions)
@@ -685,6 +742,12 @@ Tab::showTreeContextMenu(QPoint const &pos) {
 
   m_treeContextMenu->addAction(m_expandAllAction);
   m_treeContextMenu->addAction(m_collapseAllAction);
+
+  if (isTrack) {
+    m_treeContextMenu->addSeparator();
+    m_treeContextMenu->addMenu(m_modifySelectedTrackMenu);
+  }
+
   m_treeContextMenu->addSeparator();
   m_treeContextMenu->addAction(m_addAttachmentsAction);
 
@@ -876,13 +939,15 @@ Tab::isClosingOrReloadingOkIfModified(ModifiedConfirmationMode mode) {
     return true;
 
   auto modifiedPage = hasBeenModified();
-  if (!modifiedPage)
+  if (!modifiedPage && !m_tracksReordered)
     return true;
 
   auto tool = MainWindow::headerEditorTool();
   MainWindow::get()->switchToTool(tool);
   tool->showTab(*this);
-  focusPage(modifiedPage);
+
+  if (modifiedPage)
+    focusPage(modifiedPage);
 
   auto closing  = mode == ModifiedConfirmationMode::Closing;
   auto text     = closing ? QY("The file \"%1\" has been modified. Do you really want to close? All changes will be lost.")
@@ -900,4 +965,164 @@ Tab::isClosingOrReloadingOkIfModified(ModifiedConfirmationMode mode) {
   return answer == QMessageBox::Yes;
 }
 
+void
+Tab::updateTracksElementToMatchTrackOrder() {
+  auto &tracks = static_cast<libebml::EbmlMaster &>(*m_eTracks);
+
+  RemoveChildren<libmatroska::KaxTrackEntry>(tracks);
+
+  for (auto page : m_model->topLevelPages())
+    if (dynamic_cast<TrackTypePage *>(page))
+      tracks.PushElement(static_cast<TrackTypePage &>(*page).m_master);
+}
+
+void
+Tab::walkPagesOfSelectedTopLevelNode(std::function<bool(PageBase *)> worker) {
+  auto topLevelIdx = ui->elements->selectionModel()->currentIndex();
+
+  if (!topLevelIdx.isValid())
+    return;
+
+  topLevelIdx = topLevelIdx.sibling(topLevelIdx.row(), 0);
+
+  while (topLevelIdx.parent().isValid())
+    topLevelIdx = topLevelIdx.parent();
+
+  std::function<void(QModelIndex const &)> walkTree;
+  walkTree = [this, &worker, &walkTree](QModelIndex const &parentIdx) {
+    for (auto row = 0, numRows = m_model->rowCount(parentIdx); row < numRows; ++row) {
+      auto idx  = m_model->index(row, 0, parentIdx);
+
+      if (!worker(m_model->selectedPage(idx)))
+        return;
+
+      walkTree(idx);
+    }
+  };
+
+  walkTree(topLevelIdx);
+}
+
+void
+Tab::toggleSpecificTrackFlag(unsigned int wantedId) {
+  walkPagesOfSelectedTopLevelNode([wantedId](auto *pageBase) -> bool {
+    auto page = dynamic_cast<BoolValuePage *>(pageBase);
+
+    if (page && (page->m_callbacks.ClassId().GetValue() == wantedId)) {
+      page->toggleFlag();
+      return false;
+    }
+
+    return true;
+  });
+
+  updateSelectedTopLevelPageModelItems();
+}
+
+void
+Tab::toggleTrackFlag() {
+  auto action = dynamic_cast<QAction *>(sender());
+
+  if (action)
+    toggleSpecificTrackFlag(action->data().toUInt());
+}
+
+void
+Tab::changeTrackLanguage(QString const &formattedLanguage,
+                         QString const &trackName) {
+  auto language = mtx::bcp47::language_c::parse(to_utf8(formattedLanguage));
+  if (!language.is_valid())
+    return;
+
+  auto languageFound  = false,
+    languageIETFFound = false;
+
+  walkPagesOfSelectedTopLevelNode([&language, &languageFound, &languageIETFFound](auto *pageBase) -> bool {
+    auto languagePage = dynamic_cast<LanguageValuePage *>(pageBase);
+
+    if (languagePage) {
+      languagePage->setLanguage(language);
+      languageFound = true;
+    }
+
+    auto languageIETFPage = dynamic_cast<LanguageIETFValuePage *>(pageBase);
+
+    if (languageIETFPage) {
+      languageIETFPage->setLanguage(language);
+      languageIETFFound = true;
+    }
+
+    return !languageFound || !languageIETFFound;
+  });
+
+  if (trackName.isEmpty()) {
+    updateSelectedTopLevelPageModelItems();
+    return;
+  }
+
+  walkPagesOfSelectedTopLevelNode([&trackName](auto *pageBase) -> bool {
+    auto trackNamePage = dynamic_cast<TrackNamePage *>(pageBase);
+
+    if (!trackNamePage)
+      return true;
+
+    trackNamePage->setString(trackName);
+
+    return false;
+  });
+
+  updateSelectedTopLevelPageModelItems();
+}
+
+void
+Tab::moveElementUpOrDown(bool up) {
+  auto focus        = App::instance()->focusWidget();
+  auto selectedIdx  = ui->elements->selectionModel()->currentIndex();
+  auto selectedPage = m_model->selectedPage(selectedIdx);
+
+  if (!selectedIdx.isValid() || !selectedPage)
+    return;
+
+  auto idxToMove = m_model->trackOrAttachedFileIndexForSelectedIndex(selectedIdx);
+  if (!idxToMove.isValid())
+    return;
+
+  auto wasExpanded = ui->elements->isExpanded(selectedIdx);
+
+  m_model->moveElementUpOrDown(idxToMove, up);
+
+  auto newIdx = m_model->indexFromPage(selectedPage);
+
+  ui->elements->setExpanded(newIdx, wasExpanded);
+
+  auto expandIdx = newIdx.parent();
+
+  while (expandIdx.isValid()) {
+    ui->elements->setExpanded(expandIdx, true);
+    expandIdx = expandIdx.parent();
+  }
+
+  auto selection = QItemSelection{newIdx, newIdx.sibling(newIdx.row(), m_model->columnCount() - 1)};
+  ui->elements->selectionModel()->setCurrentIndex(newIdx, QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Current);
+  ui->elements->selectionModel()->select(selection, QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Current);
+
+  if (focus)
+    focus->setFocus();
+}
+
+void
+Tab::updateSelectedTopLevelPageModelItems() {
+  auto topLevelIdx = ui->elements->selectionModel()->currentIndex();
+
+  if (!topLevelIdx.isValid())
+    return;
+
+  while (topLevelIdx.parent().isValid())
+    topLevelIdx = topLevelIdx.parent();
+
+  auto page = m_model->selectedPage(topLevelIdx);
+  if (dynamic_cast<TrackTypePage *>(page))
+    static_cast<TrackTypePage *>(page)->updateModelItems();
+}
+
 }
